/*
 *
 *	Utility for reporting quota to user
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <mntent.h>

#include <linux/quota.h>

#include "common.h"
#include "quotasys.h"
#ifdef RPC
#include <rpc/rpc.h>
#include "rquota.h"
#endif

#define FL_VERBOSE 1
#define FL_QUIET 2

#define PF_HEADER 1

#define MAXMSGLEN 256	/* Length of message */

int flags, print_flags;
int type;
int nameind, limit_exceed;

void print_usage_exit(char *cmdname)
{
	printf("Utility for reporting quotas to users.\nUsage: %s [-vq] [[-ug] | -u[ user] | -g[ group]]\n", cmdname);
	printf("Bugs to %s\n", MY_EMAIL);
	exit(1);
}

void parse_options(int argcnt, char **argstr)
{
	char *slash = strrchr(argstr[0], '/'), cmdname[MAXFILENAMELEN];
	int ret;

	if (!slash)
		slash = argstr[0];
	else
		slash++;
	sstrncpy(cmdname, slash, MAXFILENAMELEN);

	while ((ret = getopt(argcnt, argstr, "Vhvqug")) != EOF) {
		switch (ret) {
			case 'h':
			case '?':
				print_usage_exit(cmdname);
			case 'V':
				version();
				exit(0);
			case 'q':
				flags |= FL_QUIET;
				break;
			case 'v':
				flags |= FL_VERBOSE;
				break;
			case 'u':
				type |= QS_USRQUOTA;
				break;
			case 'g':
				type |= QS_GRPQUOTA;
				break;
		}
	}
	if (!(type & (QS_USRQUOTA | QS_GRPQUOTA)))
		type |= QS_USRQUOTA;
	if (flags & FL_VERBOSE && flags & FL_QUIET)
		flags &= ~FL_VERBOSE;
	if (type & QS_USRQUOTA && type & QS_GRPQUOTA && optind < argcnt) {
		fputs("Bad number of arguments.\n", stderr);
		print_usage_exit(cmdname);
	}
	nameind = optind;
}

void print_head(int type, qid_t id)
{
	char name[MAXNAMELEN];

	if (type & QS_USRQUOTA) {
		uid2name(id, name);
		printf("Disk quotas for user %s(%u): ", name, (uint)id);
	}
	else {
		gid2group(id, name);
		printf("Disk quotas for group %s(%u): ", name, (uint)id);
	}
}

void print_title(void)
{
	printf("%15s%8s %7s%8s%8s%8s %7s%8s%8s\n","Filesystem","blocks","quota","limit","grace","files",
	  "quota","limit","grace");
}

char overlim(uint usage, uint softlim, uint hardlim)
{
	if ((softlim && usage > softlim) || (hardlim && usage > hardlim))
		return '*';
	return ' ';
}

#ifdef RPC
/* Convert network format to our format */
void rquota2dqblk(struct mem_dqblk *dqb, rquota *rq)
{
	dqb->dqb_bhardlimit = rq->rq_bhardlimit;
	dqb->dqb_bsoftlimit = rq->rq_bsoftlimit;
	dqb->dqb_curspace = rq->rq_curblocks << QUOTABLOCK_BITS;
	dqb->dqb_ihardlimit = rq->rq_fhardlimit;
	dqb->dqb_isoftlimit = rq->rq_fsoftlimit;
	dqb->dqb_curinodes = rq->rq_curfiles;
	dqb->dqb_btime = rq->rq_btimeleft;
	dqb->dqb_itime = rq->rq_ftimeleft;
}

/*
 * Collect the requested quota information from a remote host.
 */
void rpc_rquota(char *fsname, qid_t id, int type, struct mem_dqblk *dqb)
{
	CLIENT *clnt;
	getquota_rslt *result;
	union {
		getquota_args arg;
		ext_getquota_args ext_arg;
	} args;
	char *host, *pathname;
	struct timeval timeout = { 2, 0};

	/*
	 * Initialize with NULL.
	 */
	memset(dqb, 0, sizeof(struct mem_dqblk));

	/*
	 * Convert host:pathname to seperate host and pathname.
	 */
	host = fsname;
	/*
	 * Strip off pathname on nfs mounted dir. Ignore entries of any
 	 * automounter.
	 */
	if (!(pathname = strchr(fsname, ':')) || *(pathname + 1) == '(')
		return;
	*pathname++ = '\0';

	/*
	 * First try EXT_RQUOTAPROG
	 */
	args.ext_arg.gqa_pathp = pathname;
	args.ext_arg.gqa_id = id;
	args.ext_arg.gqa_type = user2kernquota(type);

	if ((clnt = clnt_create(host, RQUOTAPROG, EXT_RQUOTAVERS, "udp")) != NULL) {
		/*
		 * Initialize unix authentication
		 */
		clnt->cl_auth = authunix_create_default();
		clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout);

		result = rquotaproc_getquota_2(&args.ext_arg, clnt);
		if (result && result->status == Q_OK)
			rquota2dqblk(dqb, &result->getquota_rslt_u.gqr_rquota);
		auth_destroy(clnt->cl_auth);
		clnt_destroy(clnt);
	} else
		result = NULL;

	if (!result || !result->status) {
		if (type & QS_USRQUOTA) {
			/*
			 * Try RQUOTAPROG because server doesn't seem to understand EXT_RQUOTAPROG
			 */
			args.arg.gqa_pathp = pathname;
			args.arg.gqa_uid = id;

			if ((clnt = clnt_create(host, RQUOTAPROG, RQUOTAVERS, "udp"))) {
				/*
				 * Initialize unix authentication
				 */
				clnt->cl_auth = authunix_create_default();
				clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout);

				result = rquotaproc_getquota_1(&args.arg, clnt);
				if (result && result->status == Q_OK)
					rquota2dqblk(dqb, &result->getquota_rslt_u.gqr_rquota);
				auth_destroy(clnt->cl_auth);
				clnt_destroy(clnt);
			}
		}
	}
}

#endif	/* RPC */

void print(char *mnt, char *dev, int type, qid_t id, int nfs)
{
	struct mem_dqblk blk;
	char msg[MAXMSGLEN];
	int quota_set, usage_set, err_occured = 0;

	msg[0] = 0;
	if (!nfs) {
		if (quotactl(QCMD(Q_GETQUOTA, user2kernquota(type)), dev, id, (void *)&blk) < 0) {
			err_occured = 1;
			if (errno == EPERM) {
				strcpy(msg, "Permission denied");
				goto print_it;
			}
			else if (errno == ESRCH || errno == ENOSYS) {
				strcpy(msg, "None");
				goto print_it;
			}
		}
	}
	else {
#ifdef RPC
		rpc_rquota(mnt, id, type, &blk);
#else
		return;
#endif
	}
	quota_set = blk.dqb_bsoftlimit || blk.dqb_bhardlimit || blk.dqb_isoftlimit || blk.dqb_ihardlimit;
	usage_set = blk.dqb_curspace || blk.dqb_curinodes;
	/* We never inform about fake quotas... */
	if (!quota_set && !usage_set)
		return;
	/* If there's no usage or no quotas, inform only in verbose mode */
	if ((!usage_set || !quota_set) && !(flags & FL_VERBOSE))
		return;
	if ((blk.dqb_bhardlimit && toqb(blk.dqb_curspace) > blk.dqb_bhardlimit) ||
	    (blk.dqb_bsoftlimit && toqb(blk.dqb_curspace) > blk.dqb_bsoftlimit) ||
	    (blk.dqb_ihardlimit && blk.dqb_curinodes > blk.dqb_ihardlimit) ||
	    (blk.dqb_isoftlimit && blk.dqb_curinodes > blk.dqb_isoftlimit))
		limit_exceed = 1;	/* Limit was exceeded */
	if (flags & FL_QUIET) {
		if (toqb(blk.dqb_curspace) > blk.dqb_bhardlimit && blk.dqb_bhardlimit)
			strcpy(msg, "Over block limit");
		else if (toqb(blk.dqb_curspace) > blk.dqb_bsoftlimit && blk.dqb_bsoftlimit)
			strcpy(msg, "Over block quota");
		if (blk.dqb_curinodes > blk.dqb_ihardlimit && blk.dqb_ihardlimit) {
			if (msg[0])
				strcat(msg, ", ");
			strcat(msg, "Over file limit");
		}
		else if (blk.dqb_curinodes > blk.dqb_isoftlimit && blk.dqb_isoftlimit) {
			if (msg[0])
				strcat(msg, ", ");
			strcat(msg, "Over file quota");
		}
		if (!msg[0])	/* Nothing to print - we are quiet */
			return;
	}
print_it:
	if (!(print_flags & PF_HEADER)) {
		print_head(type, id);
		putchar('\n');
		if (!(flags & FL_QUIET))
			print_title();
		print_flags |= PF_HEADER;
	}
	printf("%15s", dev);
	if (flags & FL_QUIET || err_occured)
		printf(": %s\n", msg);
	else {
		char buf[MAXTIMELEN];
		difftime2str(blk.dqb_btime, buf);
		printf("%8u%c%7u%7u%8s", (uint)toqb(blk.dqb_curspace), overlim(qb2kb(toqb(blk.dqb_curspace)), blk.dqb_bsoftlimit, blk.dqb_bhardlimit), (uint)qb2kb(blk.dqb_bsoftlimit), (uint)qb2kb(blk.dqb_bhardlimit), buf);
		difftime2str(blk.dqb_itime, buf);
		printf("%8u%c%7u%7u%8s\n", (uint)blk.dqb_curinodes, overlim(blk.dqb_curinodes, blk.dqb_isoftlimit, blk.dqb_ihardlimit), (uint)blk.dqb_isoftlimit, (uint)blk.dqb_ihardlimit, buf);
	}
}

void report(qid_t id, int type)
{
	char mnt[MAXFILENAMELEN], dev[MAXDEVLEN], fstype[MAXFSTYPELEN];
	int nfs;

	print_flags &= ~PF_HEADER;
	init_mount(NULL, 0);
	while (next_mount(type, dev, mnt, fstype)) {
		nfs = !strcmp(fstype, MNTTYPE_NFS);
		print(mnt, dev, type, id, nfs);
	}
	free_mount();
	if (!(flags & FL_QUIET) && !(print_flags & PF_HEADER)) {
		print_head(type, id);
		puts("None");
	}
}

void reportall(int cnt, char **names)
{
 	int i;

	if (cnt)
		for (i = 0; i < cnt; i++) {
			if (type & QS_USRQUOTA)
				report(name2uid(names[i]), QS_USRQUOTA);
			else
				report(group2gid(names[i]), QS_GRPQUOTA);
			
		}
	else {
		if (type & QS_USRQUOTA)
			report(getuid(), QS_USRQUOTA);
		if (type & QS_GRPQUOTA) {
			int groups = getgroups(0, NULL);
			gid_t *grp = smalloc(sizeof(gid_t) * groups);
			getgroups(groups, grp);
			for (i = 0; i < groups; i++)
				report(grp[i], QS_GRPQUOTA);
		}
	}
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);
	reportall(argcnt - nameind, argstr + nameind);

	return limit_exceed? 9 : 0;
}

